@@ -1,6 +1,6 @@
 ;;; wisi-run-indent-test.el --- utils for automating indentation and casing tests
 ;;
-;; Copyright (C) 2018 - 2019  Free Software Foundation, Inc.
+;; Copyright (C) 2018 - 2020  Free Software Foundation, Inc.
 ;;
 ;; This file is part of GNU Emacs.
 ;;
@@ -18,6 +18,7 @@
 ;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
 
 (require 'wisi-tests)
+(require 'wisi-prj)
 
 ;; user can set these to t in an EMACSCMD
 (defvar skip-cmds nil)
@@ -170,6 +171,11 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
   (setq indent-tabs-mode nil)
   (setq jit-lock-context-time 0.0);; for test-face
 
+  ;; Test files use wisi-prj-select-cached to parse and select a project file.
+  (setq project-find-functions (list #'wisi-prj-current-cached))
+  (setq xref-backend-functions (list #'wisi-prj-xref-backend))
+
+
   (let ((error-count 0)
 	(test-buffer (current-buffer))
 	cmd-line
@@ -185,6 +191,11 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
     ;;    of the previous EMACSCMD, and the test fails if they don't
     ;;    match.
     ;;
+    ;; EMACSRESULT_START:<first list element>
+    ;; EMACSRESULT_ADD:  <list element>
+    ;; EMACSRESULT_FINISH:
+    ;;    build a list, compare it to the result of the previous EMACSCMD.
+    ;;
     ;; EMACS_SKIP_UNLESS: <form>
     ;;   skip entire test if form evals nil
     ;;
@@ -223,7 +234,6 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
 	(looking-at ".*$")
 	(setq expected-result (save-excursion (end-of-line 1) (eval (car (read-from-string (match-string 0))))))
 	(unless (equal expected-result last-result)
-	  (when debug-on-error (debug))
 	  (setq error-count (1+ error-count))
 	  (message
 	   (concat
@@ -234,6 +244,43 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
 		    expected-result)
 	    ))))
 
+       ((string= (match-string 1) "RESULT_START")
+	(looking-at ".*$")
+	(setq expected-result (list (save-excursion (end-of-line 1) (eval (car (read-from-string (match-string 0))))))))
+
+       ((string= (match-string 1) "RESULT_ADD")
+	(looking-at ".*$")
+	(let ((val (save-excursion (end-of-line 1)
+				   (eval (car (read-from-string (match-string 0)))))))
+	  (when val
+	    (setq expected-result (append expected-result (list val))))))
+
+       ((string= (match-string 1) "RESULT_FINISH")
+	(unless (equal (length expected-result) (length last-result))
+	  (setq error-count (1+ error-count))
+	  (message
+	   (concat
+	    (format "error: %s:%d:\n" (buffer-file-name) (line-number-at-pos))
+	    (format "Length of result of '%s' does not match.\nGot    '%s',\nexpect '%s'"
+		    last-cmd
+		    (length last-result)
+		    (length expected-result)))))
+
+	(let ((i 0))
+	  (while (< i (length expected-result))
+	    (unless (equal (nth i expected-result) (nth i last-result))
+	      (setq error-count (1+ error-count))
+	      (message
+	       (concat
+		(format "error: %s:%d:\n" (buffer-file-name) (line-number-at-pos))
+		(format "Nth (%d) result of '%s' does not match.\nGot    '%s',\nexpect '%s'"
+			i
+			last-cmd
+			(nth i last-result)
+			(nth i expected-result))
+		)))
+	    (setq i (1+ i)))))
+
        ((string= (match-string 1) "_SKIP_UNLESS")
 	(looking-at ".*$")
 	(unless (eval (car (read-from-string (match-string 0))))
@@ -261,7 +308,7 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
        (buffer-file-name) (line-number-at-pos (point)) error-count))
     )
 
-  (when (not skip-reindent-test)
+  (unless skip-reindent-test
     ;; Reindent the buffer
     (message "indenting")
 
@@ -278,11 +325,18 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
     ;; files must be saved without any.
     (delete-trailing-whitespace)
     )
+
+  (when (and wisi-auto-case (not skip-recase-test))
+    (message "casing")
+    (wisi-case-adjust-buffer))
   )
 
 (defun run-test (file-name)
   "Run an indentation and casing test on FILE-NAME."
   (interactive "f")
+
+  (package-initialize) ;; for uniquify-files
+
   ;; we'd like to run emacs from a makefile as:
   ;;
   ;; emacs -Q --batch -l runtest.el -f run-test-here <filename>
@@ -304,14 +358,11 @@ Each item is a list (ACTION PARSE-BEGIN PARSE-END EDIT-BEGIN)")
   ;; that rely on font-lock do explicitly.
   (setq font-lock-support-mode nil)
 
+  (setq xref-prompt-for-identifier nil)
+
   (let ((dir default-directory))
     (find-file file-name) ;; sets default-directory
 
-    (when (eq major-mode 'fundamental-mode)
-      ;; Running a grammar in test/wisi
-      (add-to-list 'load-path (expand-file-name "."))
-      (wisi-tests-setup (file-name-sans-extension (file-name-nondirectory file-name))))
-
     (run-test-here)
 
     (unless skip-write
